@@ -3,6 +3,5 @@
 
 import affinegap
 from haversine import compareLatLong
-from jaccard import compareJaccard
-from cosine import createCosineSimilarity, CosineSimilarity
+from cosine import CosineTextSimilarity, CosineSetSimilarity
 from categorical import CategoricalComparator
